/* book pages 147-150 */
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define NEURONS     10
#define LINKS        8
#define BUFSIZE   4192
#define SIGMOIDAL 0
#define IDENTITY  1

typedef struct Neuron {
     int neuronId;
     int layerId;
     double threshold;
     int functionType;
     double inputs [LINKS];
     double weights [LINKS];
     } Neuron;

Neuron sourceLayer [NEURONS], targetLayer [NEURONS];
double inputVector [LINKS] ={0.250, 0.375, 0.020, 0.0500,
                             0.975, 0.543, 0.125, 0.0075};
void InitLayer (void) {
    int i, j;
    for (i=0;i<NEURONS;i++) {
         sourceLayer[i].layerId = 1;
         sourceLayer[i].neuronId = i;
         sourceLayer[i].threshold = (double)rand()/RAND_MAX;
         sourceLayer[i].functionType = SIGMOIDAL;
         for (j=0;j<LINKS;j++) {
              sourceLayer[i].inputs[j] = inputVector[j];
              sourceLayer[i].weights[j] =
                    (double)rand()/RAND_MAX; }}}
int main (int argc, char ** argv) {
     int i, j, rank, size;
     double sum, output=0;
     int packPosition;
     char packBuffer [BUFSIZE];
     MPI_Status status;
     MPI_Init (&argc, &argv);
     MPI_Comm_rank (MPI_COMM_WORLD, &rank);
     MPI_Comm_size (MPI_COMM_WORLD, &size);

     if (size<2) {
         if (rank==0)
             printf ("Invalid process number. Aborting...\n");
         MPI_Finalize();
         return (-1); }

     InitLayer ();
     if (rank==0) {
         packPosition = 0;
         for (i=0;i<NEURONS;i++) {
              MPI_Pack (&sourceLayer[i].neuronId,1,MPI_INT,
                  packBuffer,BUFSIZE,&packPosition,
                  MPI_COMM_WORLD);
              MPI_Pack (&sourceLayer[i].layerId,1,MPI_INT,
                  packBuffer, BUFSIZE,&packPosition,
                  MPI_COMM_WORLD);
              MPI_Pack (&sourceLayer[i].threshold,1,MPI_DOUBLE,
                  packBuffer,BUFSIZE,&packPosition,
                  MPI_COMM_WORLD);
              MPI_Pack (&sourceLayer[i].functionType,1,MPI_INT,
                  packBuffer,BUFSIZE,&packPosition,
                  MPI_COMM_WORLD);
              MPI_Pack (sourceLayer[i].inputs,LINKS,MPI_DOUBLE,
                  packBuffer,BUFSIZE,&packPosition,
                  MPI_COMM_WORLD);
              MPI_Pack (sourceLayer[i].weights,LINKS,MPI_DOUBLE,
                  packBuffer,BUFSIZE,&packPosition,
                  MPI_COMM_WORLD); }

         MPI_Send (packBuffer,packPosition,MPI_PACKED,
                   1,0,MPI_COMM_WORLD); }
     else {
         packPosition = 0;
         MPI_Recv (packBuffer,BUFSIZE,MPI_PACKED,0,0,
                   MPI_COMM_WORLD,&status);
         for (i=0;i<NEURONS;i++) {
              MPI_Unpack (packBuffer,BUFSIZE,&packPosition,
                   &targetLayer[i].neuronId,1,MPI_INT,
                   MPI_COMM_WORLD);
              MPI_Unpack (packBuffer,BUFSIZE,&packPosition,
                   &targetLayer[i].layerId,1,MPI_INT,
                   MPI_COMM_WORLD);
              MPI_Unpack (packBuffer, BUFSIZE, &packPosition,
                   &targetLayer[i].threshold,1,MPI_DOUBLE,
                   MPI_COMM_WORLD);
              MPI_Unpack (packBuffer,BUFSIZE,&packPosition,
                   &targetLayer[i].functionType,1,MPI_INT,
                   MPI_COMM_WORLD);
              MPI_Unpack (packBuffer,BUFSIZE,&packPosition,
                   targetLayer[i].inputs,LINKS,MPI_DOUBLE,
                   MPI_COMM_WORLD);
              MPI_Unpack (packBuffer, BUFSIZE, &packPosition,
                   targetLayer[i].weights,LINKS,MPI_DOUBLE,
                   MPI_COMM_WORLD); }

         printf ("\nLayer Configuration\n");
         printf ("-------------------\n\n");
         for (i=0;i<NEURONS;i++) {
              printf ("Neuron %02d --> NeuronId  = %02d\n",
                      i,targetLayer[i].neuronId);
              printf ("Neuron %02d --> LayerId   = %02d\n",
                      i,targetLayer[i].layerId);
              printf ("Neuron %02d --> Threshold = %.3f\n",
                      i,targetLayer[i].threshold);
              if (targetLayer[i].functionType==SIGMOIDAL)
                  printf ("Neuron %02d-->
                      Function=SIGMOIDAL\n", i);
              if (targetLayer[i].functionType==IDENTITY)
                  printf ("Neuron %02d-->
                      Function=IDENTITY\n", i);
              printf ("Neuron %02d --> Inputs = ", i);
              for (j=0;j<LINKS;j++)
                   printf ("%.3f ", targetLayer[i].inputs[j]);
              printf ("\n");
              printf ("Neuron %02d --> Weights = ", i);
              for (j=0;j<LINKS;j++)
                   printf ("%.3f ", targetLayer[i].weights[j]);
              printf ("\n\n"); }
         printf ("\nLayer Output\n");
         printf ("-------------------\n\n");

         for (i=0;i<NEURONS;i++) {
              sum = 0.0, output = 0.0;
              for (j=0;j<LINKS;j++)
                   sum += targetLayer[i].inputs[j]*
                          targetLayer[i].weights[j];
              sum -= targetLayer[i].threshold;
              if (targetLayer[i].functionType==IDENTITY)
                  output = sum;
              if (targetLayer[i].functionType==SIGMOIDAL)
                  output = 1/(1+exp(-sum));
              printf ("Neuron %02d --> Output = %.6f\n",
                      i, output); }}

     MPI_Finalize ();
     return (0); }
