/* book pages 283-284 */
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

static int mCast_key;

int mCast_Delete_Func (MPI_Comm comm, int keyVal,
          void * attrVal, void * extraState) {
    return (MPI_Comm_free((MPI_Comm *)&attrVal)); }

void mCast_Init (void) {
    MPI_Keyval_create (MPI_NULL_COPY_FN, mCast_Delete_Func,
                       &mCast_key, NULL); }

void mCast (void * buffer, int count, MPI_Datatype type,
            int isRoot, MPI_Comm comm) {
    int rank, size, flag, numLeaves;
    int child, childLeaves;
    MPI_Status status;
    MPI_Comm pComm;
    int attrVal;

    MPI_Attr_get (comm, mCast_key, &attrVal, &flag);

    if (flag) {
        pComm = (MPI_Comm)attrVal; }
    else {
        MPI_Comm_dup (comm, &pComm);
        MPI_Attr_put (comm, mCast_key, (void *)pComm); }

    MPI_Comm_rank (pComm, &rank);
    MPI_Comm_size (pComm, &size);

    if (isRoot)  numLeaves = size-1;
    else {
        MPI_Recv (&numLeaves, 1, MPI_INT, MPI_ANY_SOURCE,
                  0, pComm, &status);
        MPI_Recv (buffer, count, type, MPI_ANY_SOURCE,
                  0, pComm, &status);
    while (numLeaves>0) {
        child = (rank+(numLeaves+1)/2)%size;
        childLeaves = numLeaves/2;
        MPI_Send (&childLeaves, 1, MPI_INT, child, 0, pComm);
        MPI_Send (buffer, count, type, child, 0, pComm);
        printf ("Message sent from process %02d to
                 process %02d\n", rank, child);
        numLeaves -= (childLeaves+1); }}

int main (int argc, char ** argv) {
    int rank, size, a=10;
    MPI_Status status;
    MPI_Init (&argc, &argv);
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    MPI_Comm_size (MPI_COMM_WORLD, &size);

    if (rank==2) {
        MPI_Send (&a, 1, MPI_INT, 1, 0, MPI_COMM_WORLD);
        printf ("MPI_Send called from process 2\n"); }

    mCast_Init();
    mCast (&a, 1, MPI_INT, (rank==0), MPI_COMM_WORLD);

    if (rank==1) {
        MPI_Recv (&a, 1, MPI_INT, MPI_ANY_SOURCE, 0,
                  MPI_COMM_WORLD, &status);
        printf ("MPI_Recv called from process 1\n"); }

    MPI_Finalize ();
    return (0); }
