#include <mpi.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>

#define LENGTH    32
#define MESSAGES  10
#define LOGLEN     5
#define PRINT      8
#define DISCONNECT 100
#define CALCULATE  200

#define MAXHOSTNAMELEN 128

int main(int argc, char **argv) {
    MPI_Comm newComm;
    char portName [MPI_MAX_PORT_NAME];
    char * serviceName = "fftService";
    int rank, size, msgs=0, done=0, counter;
    double message [256];
    char hostname[MAXHOSTNAMELEN];

    gethostname(hostname, MAXHOSTNAMELEN);

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    MPI_Lookup_name(serviceName, MPI_INFO_NULL, portName);

    printf("Client %0d ==> Service Name %s, Port Name %s\n",
            rank, serviceName, portName);

    MPI_Comm_connect(portName, MPI_INFO_NULL, 0,
                     MPI_COMM_WORLD, &newComm);

    while (!done) {
       if (msgs==MESSAGES) done=1;
       for (counter=0;counter<LENGTH;counter++)
            message[counter]=(double)rand()/RAND_MAX;
       MPI_Send(message,LENGTH, MPI_DOUBLE, 0,
                CALCULATE, newComm );
       printf ("Message [%03d] Data: ", msgs);
       for (counter=0;counter<PRINT;counter++)
            printf ("%.3f ", message[counter]);
       printf ("\n");
       msgs++; }

    MPI_Comm_disconnect(&newcomm);
    MPI_Finalize();
    return 0; }

// Source File Server.c

#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <math.h>

#define LENGTH 32
#define LOGLEN  5
#define PRINT   5

#define DISCONNECT 100
#define CALCULATE  200

#define SWAP(a,b)  tempr=(a);(a)=(b);(b)=tempr
#define PI        (4.0)*atan(1.0)
#define PRINT      6

MPI_Comm newComm;
char portName [MPI_MAX_PORT_NAME];
char * serviceName = "fftService";

int bit_swap (int i, int nu) {
    int ib=0, i1, i2;
    for(i1=0;i1!=nu;i1++) {
         i2 = i/2;
         ib = ib*2+i-2*i2;
         i = i2;}
    return(ib); }

void swap(double x1, double x2)  {
    double temp_x;
    temp_x = x1;
    x1 = x2;
    x2 = temp_x; }


void fft (double * real_data, double * imag_data,
          int n_pts, int nu, int inv) {
    int n2, j, l, i, ib, k, k1, k2, sgn;
    double tr, ti, arg, nu1;
    double c, s;
    n2 = n_pts / 2;
    nu1 = nu - 1.0;
    k = 0;
    sgn = inv ? -1 : 1;
    for (l=0;l!=nu;l++) {
      do {
          for (i=0;i!=n2;i++) {
             j = (int)(k/(pow(2.0,nu1)));
             ib = bit_swap( j, nu );
             arg = 2.0 * PI * ib / n_pts;
             c = cos( arg );
             s = sgn * sin( arg );
             k1 = k;
             k2 = k1 + n2;
             tr = *(real_data+k2) * c + *(imag_data+k2) * s;
             ti = *(imag_data+k2) * c - *(real_data+k2) * s;
             *(real_data+k2) = *(real_data+k1) - tr;
             *(imag_data+k2) = *(imag_data+k1) - ti;
             *(real_data+k1) = *(real_data+k1) + tr;
             *(imag_data+k1) = *(imag_data+k1) + ti;
             k++; }
          k +=  n2;
          } while (k<n_pts-1);
      k = 0;
      nu1 -= 1.0;
      n2 /= 2; }
    for (k=0;k!=n_pts;k++) {
         ib = bit_swap( k, nu );
         if (ib>k) {
         swap(*(real_data+k), *(real_data+ib));
         swap(*(imag_data+k), *(imag_data+ib));}}
    if (inv)
        for (k=0;k!=n_pts;k++) {
             *(real_data+k) /= n_pts;
             *(imag_data+k) /= n_pts; }}


int main (int argc, char **argv) {
    MPI_Status status;
    int rank, size, counter;
    double message [LENGTH];
    double realData[LENGTH];
    double imagData[LENGTH];

    printf ("Server:> server is running, hit Ctrl-C
            to terminate\n");

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    MPI_Open_port(MPI_INFO_NULL, portName);
    printf ("Server:> port name %s is available for requests\n",
            portName);

    MPI_Publish_name(serviceName, MPI_INFO_NULL, portName);
    printf ("Server:> service name %s has been published\n",
           serviceName);
    signal(SIGINT, quit);

    while (1) {
       printf("Server:> accepting connections ... \n");
       MPI_Comm_accept(portName, MPI_INFO_NULL, 0,
                       MPI_COMM_WORLD, &newComm);
       while (1) {
            MPI_Recv (message, LENGTH, MPI_DOUBLE,
                      MPI_ANY_SOURCE, MPI_ANY_TAG,
                      newComm, &status);
            if (status.MPI_TAG==DISCONNECT) {
                printf("Server:> disconnection request
                       from client\n");
                MPI_Comm_disconnect(&newComm);
                break; }
            if (status.MPI_TAG==CALCULATE){
                for (counter=0;counter<LENGTH;counter++) {
                     realData[counter]=message[counter];
                     imagData[counter]=0.0; }
                fft (realData, imagData, LENGTH, LOGLEN, 1);
                printf ("REAL DATA for Client %02d: ",
                         status.MPI_SOURCE);
                for (counter=0;counter<PRINT;counter++)
                     printf ("%+.3f  ", realData[counter]);
                printf ("\n");
                printf ("IMAG DATA for Client %02d: ",
                        status.MPI_SOURCE);
                for (counter=0;counter<PRINT;counter++)
                     printf ("%+.3f  ", imagData[counter]);
                printf ("\n"); }}}}