/* book pages 663-665 */

#include <mpi.h>
#include <mpe.h>
#include <math.h>
#include <stdio.h>

double f (double a) {
    return (4.0 / (1.0 + a*a)); }

int main(int argc, char **argv) {
    int  n, rank, size, i, j, namelen;
    double PI25DT = 3.141592653589793238462643;
    double mypi, pi, h, sum, x;
    double startwtime = 0.0, endwtime;
    int startBcast, endBcast, startCompute, endCompute;
    int startReduce, endReduce, startSync, endSync;
    char processor_name[MPI_MAX_PROCESSOR_NAME];

    MPI_Init (&argc, &argv);
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    MPI_Comm_size (MPI_COMM_WORLD, &size);

    MPI_Get_processor_name(processor_name,&namelen);
    fprintf(stderr,"Process %d running on %s\n",
            rank, processor_name);

    MPE_Init_log();

    startBcast = MPE_Log_get_event_number();
    endBcast = MPE_Log_get_event_number();
    startCompute = MPE_Log_get_event_number();
    endCompute = MPE_Log_get_event_number();
    startReduce = MPE_Log_get_event_number();
    endReduce = MPE_Log_get_event_number();
    startSync = MPE_Log_get_event_number();
    endSync = MPE_Log_get_event_number();

    if (rank == 0) {
        MPE_Describe_state(startBcast, endBcast,
            "Broadcast", "red");
        MPE_Describe_state(startCompute, endCompute,
            "Compute", "blue");
        MPE_Describe_state(startReduce, endReduce,
            "Reduce", "green");
        MPE_Describe_state(startSync, endSync,
            "Sync", "orange");  }
        n= 1000000;
        startwtime = MPI_Wtime(); }
    MPI_Barrier(MPI_COMM_WORLD);
    MPE_Start_log();

    for (j = 0; j < 5; j++) {
         MPE_Log_event(startBcast, 0, "Start Broadcast");
         MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);
         MPE_Log_event(endBcast, 0, "End Broadcast");

         MPE_Log_event(startSync,0,"Start Sync");
         MPI_Barrier(MPI_COMM_WORLD);
         MPE_Log_event(endSync,0,"End Sync");

         MPE_Log_event(startCompute, 0, "Start Compute");
         h  = 1.0 / (double) n;
         sum = 0.0;
         for (i = rank + 1; i <= n; i += size) {
              x = h * ((double)i - 0.5); sum += f(x); }
              mypi = h * sum;
         MPE_Log_event(endCompute, 0, "End Compute");

         MPE_Log_event(startReduce, 0, "Start Reduce");
         MPI_Reduce(&mypi, &pi, 1, MPI_DOUBLE,
                    MPI_SUM, 0, MPI_COMM_WORLD);
         MPE_Log_event(endReduce, 0, "End Reduce");  }

    MPE_Finish_log("cpilog");

    if (rank == 0) {
        endwtime = MPI_Wtime();
        printf("pi is approximately %.16f, Error is %.16f\n",
                pi, fabs(pi - PI25DT));
        printf("wall clock time = %f\n",
                endwtime-startwtime); }
    MPI_Finalize();
    return(0); }
