/* book pages 644-646 */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

int MPI_Comm_rank (MPI_Comm comm) {
    int rank, retVal = PMPI_Comm_rank (comm, &rank);
    if (retVal!=MPI_SUCCESS) {
        printf ("Error in PMPI_Comm_rank function\n");
        return (-1); }
    else return (rank); }

int MPI_Comm_size (MPI_Comm comm) {
    int size, retVal = PMPI_Comm_size (comm, &size);
    if (retVal!=MPI_SUCCESS) {
        printf ("Error in PMPI_Comm_size function\n");
        return (-1); }
    else return (size); }

int MPI_Init (int * argc, char *** argv) {
    int rank, retVal;
    double startTime, endTime;
    startTime = PMPI_Wtime();
    retVal = PMPI_Init (argc, argv);
    endTime = PMPI_Wtime();
    rank = MPI_Comm_rank (MPI_COMM_WORLD);
    printf ("Intializing MPI environment for process %02d...",
             rank);
    if (retVal!=MPI_SUCCESS) {
        printf ("Error in MPI_Init function\n");
        return (-1); }
    else {
        printf ("Start %f End %f\n", startTime, endTime);
        return (retVal); }}

int MPI_Send (void * buffer, int count, MPI_Datatype dType,
              int dest, int tag, MPI_Comm comm) {
    int retVal, rank, extent;
    double startTime, endTime;
    rank = MPI_Comm_rank (comm);
    PMPI_Type_extent (dType, &extent);
    startTime = PMPI_Wtime();
    retVal = PMPI_Send (buffer,count,dType,dest,tag,comm);
    endTime = PMPI_Wtime();
    if (retVal!=MPI_SUCCESS) {
        printf ("Error in MPI_Send function\n");
        return (-1); }
    else {
        printf ("P%d sent to P%d a message of %d bytes
                 in %.6f secs (Start %f End %f)\n",
                 rank, dest, count*extent, (endTime-startTime),
                 startTime, endTime);
    return (retVal); }}

int MPI_Recv (void * buffer, int count, MPI_Datatype dType,
        int source, int tag, MPI_Comm comm,
        MPI_Status * status) {
    int retVal, rank, extent;
    double startTime, endTime;
    rank = MPI_Comm_rank (comm);
    PMPI_Type_extent (dType, &extent);
    startTime = PMPI_Wtime();
    retVal = PMPI_Recv (buffer, count, dType, source,
                        tag, comm, status);
    endTime = PMPI_Wtime();
    if (retVal!=MPI_SUCCESS) {
        printf ("Error in MPI_Recv function\n");
        return (-1); }
    else {
        printf ("P%d received from P%d a message of %d bytes \
                 in %.6f secs (Start %f End %f)\n",
                 rank,source,count*extent,(endTime-startTime),
                 startTime, endTime);
    return (retVal); }}

int MPI_Finalize (void) {
    int retVal, rank = MPI_Comm_rank (MPI_COMM_WORLD);
    double startTime, endTime;
    printf ("MPI Environment is shutting down
            for process %02d...", rank);
    startTime = PMPI_Wtime();
    retVal =  PMPI_Finalize();
    endTime = PMPI_Wtime();
    if (retVal!=MPI_SUCCESS) {
        printf ("Error in MPI_Finalize function\n");
        return (-1); }
    else {
        printf ("Start %f End %f\n", startTime, endTime);
        return (retVal); }}

void main(int argc, char ** argv) {
    int counter, rank, size;
    double a;
    MPI_Status status;
    MPI_Init (&argc, &argv);
    rank = MPI_Comm_rank (MPI_COMM_WORLD);
    size = MPI_Comm_size (MPI_COMM_WORLD);
    if (rank==0) {
        a = 3.14159;
        for (counter=1;counter<size;counter++)
             MPI_Send (&a,1,MPI_DOUBLE,counter,0,
                 MPI_COMM_WORLD); }
    else MPI_Recv (&a,1,MPI_DOUBLE,0,0,MPI_COMM_WORLD,&status);
    MPI_Finalize(); }
