/* book pages 743-747 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <mpi.h>
#include <mpe.h>
#include <mpe_graphics.h>

#define DISP 15

void DrawPoint (MPE_XGraph graph, double x, double y, int width,
         int height, double hMin, double hMax, double vMin,
         double vMax, MPE_Color color) {
    int xZ, xM = width, x0 = 0, yZ, y0 = 0, yM = height;
    double hUnit, vUnit;

    if ((hMin>=0)&&(hMax>0)) {
         if ((vMin>=0)&&(vMax>0)) {
              xZ=x0;
              yZ=yM;
              MPE_Draw_point (graph, xM-(int)((width/
                  (hMax-hMin))*(Max-x)), yZ-(int)
                  ((yZ-y0)/((vMax)*y)),color); }
         if ((vMin<0)&&(vMax>=0)) {
              xZ=x0;
              yZ=yM-(int)((yM-y0)*fabs(vMin)/
                  (fabs(vMax)+fabs(vMin)));
              vUnit = (yM-y0)/(fabs(vMin)+fabs(vMax));
              MPE_Draw_point (graph, xM-(int)((width/
                  (hMax-hMin))*(hMax-x)), yZ-
                  (int)(y*vUnit), color); }
         if ((vMin<0)&&(vMax<=0)) {
              xZ=x0;
              yZ=y0;
              vUnit=(yM-y0)/fabs(vMin);
              MPE_Draw_point (graph, xM-(int)(((width)/
                  (hMax-hMin))*(hMax-x)),yZ-
                  (int)(y*vUnit), color); }}
    if ((hMin<=0)&&(hMax>0)) {
         if ((vMin>=0)&&(vMax>0)) {
              xZ=x0+(int)((width)*fabs(hMin)/(fabs(hMin)+
                  fabs(hMax)));
              yZ=yM;
              hUnit = (width)/(fabs(hMin)+fabs(hMax));
              MPE_Draw_point (graph, xZ+(int)(x*hUnit),
                 yZ-(int)(((yZ-y0)/(vMax))*y), color);}
         if ((vMin<0)&&(vMax>=0)) {
              hUnit = (width)/(fabs(hMin)+fabs(hMax));
              vUnit = (yM-y0)/(fabs(vMin)+fabs(vMax));
              xZ=x0+(int)((width)*fabs(hMin)/
                  (fabs(hMax)+fabs(hMin)));
              yZ=y0+(int)((yM-y0)*fabs(vMax)/
                  (fabs(vMax)+fabs(vMin)));
              MPE_Draw_point (graph, xZ+(int)(x*hUnit),
                 yZ-(int)(y*vUnit),color);}
         if ((vMin<0)&&(vMax<=0)) {
              xZ=x0+(int)((width)*fabs(hMin)/(fabs(hMax)
                  +fabs(hMin)));
              yZ=y0;
              hUnit = (width)/(fabs(hMin)+fabs(hMax));
              vUnit=(yM-y0)/fabs(vMin);
              MPE_Draw_point (graph, xZ+(int)(x*hUnit),
                 yZ-(int)(y*vUnit), color);}}
    if ((hMin<0)&&(hMax<=0)) {
         if ((vMin>=0)&&(vMax>0)) {
              xZ=xM; yZ=yM;
              hUnit=(width)/(fabs(hMin)-fabs(hMax));
              MPE_Draw_point (graph, x0+(int)(hUnit*
                  (fabs(hMin)-fabs(x))), yZ-(int)(((yZ-y0)/
                  (vMax))*y), color); }
         if ((vMin<0)&&(vMax>=0)) {
              xZ=xM;
              yZ=yM-(int)((yM-y0)*fabs(vMin)/(fabs(vMax)+
                  fabs(vMin)));
              hUnit=(width)/(fabs(hMin)-fabs(hMax));
              vUnit = (yM-y0)/(fabs(vMin)+fabs(vMax));
              MPE_Draw_point (graph, x0+(int)(hUnit*
                  (fabs(hMin)-fabs(x))),
                  yZ-(int)(y*vUnit), color); }
         if ((vMin<0)&&(vMax<=0)) {
              xZ=xM; yZ=y0;
              hUnit=(width)/(fabs(hMin)-fabs(hMax));
              vUnit=(yM-y0)/fabs(vMin);
              MPE_Draw_point (graph, x0+(int)
                  (hUnit*(fabs(hMin)-fabs(x))),
                  yZ-(int)(y*vUnit),color); }}}

int main (int argc, char ** argv) {
    int counter ,samples = 1000000;
    int rank, size, xCoord = 10, yCoord = 50;
    int width = 300, height = 300;
    MPE_XGraph graph;
    double aParam = 10.0, bParam = 28.0;
    double cParam = 2.666, deltaT = 0.0001;
    double x,y,z,x0=0.1,y0=0.1,z0=0.1;
    double hMin, hMax,  vMin, vMax;
    char key;

    MPI_Init (&argc, &argv);
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    MPI_Comm_size (MPI_COMM_WORLD, &size);

    if (size!=3) {
       if (rank==0)
           printf ("This application requires exactly
                    three processes\n");
       MPI_Finalize();
       return (-1); }
    if (rank==0) {
        hMin=-30.0; hMax=+30.0;
        vMin=-30.0; vMax=+30.0; }
    if ((rank==1)||(rank==2)) {
         hMin=-30.0; hMax=+30.0;
         vMin=+0.00; vMax=+60.0; }

    MPE_Open_graphics (&graph, MPI_COMM_SELF, NULL,
        xCoord+rank*(width+DISP), yCoord, width, height, 0);

    if (rank==0)
        MPE_Draw_string (graph, 20, 20, MPE_BLACK,
            "XY Projection");
    if (rank==1)
        MPE_Draw_string (graph, 20, 20, MPE_BLACK,
            "XZ Projection");
    if (rank==2)
        MPE_Draw_string (graph, 20, 20, MPE_BLACK,
            "YZ Projection");
    for (counter=0;counter<samples;counter++) {
         x = (1-aParam*deltaT)*x0 + aParam*deltaT*y0;
         y = (bParam-z0)*deltaT*x0 + (1-deltaT)*y0;
         z = deltaT*x0*y0 + (1-cParam*deltaT)*z0;
         if (rank==0)
             DrawPoint (graph, x, y, width, height, hMin,
                        hMax, vMin, vMax, MPE_BLACK);
         if (rank==1)
             DrawPoint (graph, x, z, width, height, hMin,
                        hMax, vMin, vMax, MPE_RED);
         if (rank==2)
             DrawPoint (graph, y, z, width, height, hMin,
                        hMax, vMin, vMax, MPE_BLUE);
         x0 = x; y0 = y; z0 = z; }

    if (rank==0) {
        printf ("Press any key to continue...\n");
        scanf ("%s", &key);
        printf ("\n"); }

    MPI_Barrier (MPI_COMM_WORLD);
    if (graph)
        MPE_Close_graphics (&graph);
    MPI_Finalize ();
    return (0); }
