/* book pages 737-739 */

#include <math.h>
#include <stdlib.h>
#include <mpi.h>
#include <mpe.h>
#include <mpe_graphics.h>

#define SAMPLES 6000
#define LOOPS   100

#define minLimit 1
#define maxLimit 4

int main (int argc, char ** argv) {
    int xCoord = 50, yCoord = 50;
    int counter, index, retVal, slice, xStart;
    int rank, size, localSamples;
    double * aVal, * yVal, interval, xVal;
    double minA, maxA, minY, maxY, step;
    double minValue, maxValue;
    int width = 1000, height = 600;
    static MPE_XGraph graph;
    char key;

    MPI_Init (&argc, &argv);
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    MPI_Comm_size (MPI_COMM_WORLD, &size);

    interval = (double)(maxLimit-minLimit)/size;
    slice = width/size;
    xStart = xCoord+rank*slice;
    localSamples = SAMPLES/size;
    minValue = minLimit+rank*interval;
    maxValue = minValue + interval;
    step = (double)(maxValue-minValue)/localSamples;
    aVal = (double *) malloc (localSamples*sizeof(double));
    if (!aVal) {
        printf ("Process (%02d) Not enough memory
                 for aVal buffer\n", rank);
        MPI_Finalize ();
        return (-1);}
    yVal = (double *) malloc (localSamples*sizeof(double));
    if (!yVal) {
        printf ("Process (%02d) Not enough memory
                 for yVal buffer\n", rank);
        MPI_Finalize ();
        return (-1);}

    for (counter=0;counter<localSamples;counter++)
         aVal[counter] = yVal[counter]=0.0;
    for (counter=0;counter<localSamples;counter++) {
         aVal [counter] = minValue+(step*counter);
         xVal = (double)rand()/RAND_MAX;
         for (index=0;index<LOOPS;index++)
              xVal = aVal[counter]*xVal*(1-xVal);
         yVal[counter]=xVal;}

    minA = aVal [0];
    for (counter=0;counter<localSamples;counter++)
         if (aVal[counter]<minA)
             minA = aVal [counter];
    maxA = aVal [0];
    for (counter=0;counter<localSamples;counter++)
         if (aVal[counter]>maxA)
                     maxA = aVal [counter];

    maxY = 1.0;  minY = 0.0;
    retVal = MPE_Open_graphics (&graph, MPI_COMM_WORLD, NULL,
                      xCoord, xCoord, width, height, 1);
    if (retVal!=MPE_SUCCESS) {
        printf ("Process (%02d) MPE_Open_graphics
                 failed to start\n",   rank);
        MPI_Finalize ();
        return (-1);}

    for (counter=0;counter<localSamples;counter++)
         MPE_Draw_point (graph, xStart+slice-
             maxA-aVal[counter])*(slice/(maxA-minA)),
             yCoord+(maxY-yVal[counter])*
            (height/(maxY-minY)), rank+1);
    MPE_Update (graph);

    if (rank==0) {
        printf ("Press any key to continue...\n");
        scanf ("%s", &key);
        printf ("\n"); }
    MPI_Barrier (MPI_COMM_WORLD);
    MPE_Close_graphics (&graph);
    MPI_Finalize ();
    free (aVal);
    free (yVal);
    return (0); }
